from typing import TYPE_CHECKING, Any, ClassVar, TypeVar

from hexbytes.main import HexBytes as BaseHexBytes
from pydantic_core.core_schema import (
    ValidationInfo,
    bytes_schema,
    with_info_before_validator_function,
)

from eth_pydantic_types.hex.base import BaseHex
from eth_pydantic_types.serializers import hex_serializer
from eth_pydantic_types.utils import (
    PadDirection,
    get_hash_examples,
    get_hash_pattern,
    validate_bytes_size,
)

if TYPE_CHECKING:
    from pydantic_core import CoreSchema
    from typing_extensions import TypeAlias


HexBytesSelf = TypeVar("HexBytesSelf", bound="HexBytes")


class HexBytes(BaseHexBytes, BaseHex):
    """
    Use when receiving ``hexbytes.HexBytes`` values. Includes
    a pydantic validator and serializer.
    """

    @classmethod
    def __get_pydantic_core_schema__(cls: type[HexBytesSelf], value, handle=None) -> "CoreSchema":
        schema = with_info_before_validator_function(cls.__eth_pydantic_validate__, bytes_schema())
        schema["serialization"] = hex_serializer
        return schema

    @classmethod
    def fromhex(cls: type[HexBytesSelf], hex_str: str) -> HexBytesSelf:
        value = hex_str[2:] if hex_str.startswith("0x") else hex_str
        return super().fromhex(value)

    @classmethod
    def __eth_pydantic_validate__(
        cls: type[HexBytesSelf],
        value: Any,
        info: ValidationInfo | None = None,
        **kwargs,
    ) -> HexBytesSelf:
        if not (pad := kwargs.pop("pad", None)):
            pad = PadDirection.LEFT if isinstance(value, int) else PadDirection.RIGHT

        return cls(cls.validate_size(HexBytes(value), pad_direction=pad))

    @classmethod
    def validate_size(
        cls: type[HexBytesSelf], value: bytes, pad_direction: PadDirection = PadDirection.LEFT
    ) -> bytes:
        return value


class BoundHexBytes(HexBytes):
    """
    Use when receiving ``hexbytes.HexBytes`` values and a specific size is required.
    Includes a pydantic validator and serializer.
    """

    size: ClassVar[int] = 32

    @classmethod
    def __get_pydantic_core_schema__(cls: type[HexBytesSelf], value, handle=None) -> "CoreSchema":
        schema = with_info_before_validator_function(
            cls.__eth_pydantic_validate__,
            bytes_schema(max_length=cls.size, min_length=cls.size),
        )
        schema["serialization"] = hex_serializer
        return schema

    @classmethod
    def validate_size(
        cls: type[HexBytesSelf], value: bytes, pad_direction: PadDirection = PadDirection.LEFT
    ) -> bytes:
        str_size = cls.size * 2
        cls.schema_pattern = get_hash_pattern(str_size)
        cls.schema_examples = get_hash_examples(str_size)
        return validate_bytes_size(value, cls.size, pad_direction=pad_direction)


class HexBytes20(BoundHexBytes):
    size: ClassVar[int] = 20


HexBytes32: "TypeAlias" = BoundHexBytes



from typing_extensions import assert_type

from eth_pydantic_types.hex.bytes import HexBytes32


def test_hexbytes_fromhex():
    x = HexBytes32.fromhex("00" * 32)
    assert_type(x, HexBytes32)
    assert isinstance(x, HexBytes32)



from functools import cached_property
from typing import TYPE_CHECKING, Annotated, Any, ClassVar

from cchecksum import to_checksum_address
from eth_typing import ChecksumAddress
from pydantic_core.core_schema import (
    ValidationInfo,
    str_schema,
    with_info_before_validator_function,
)

from eth_pydantic_types.hex import HexStr20
from eth_pydantic_types.utils import (
    PadDirection,
)

if TYPE_CHECKING:
    from pydantic_core import CoreSchema

ADDRESS_PATTERN = "^0x[a-fA-F0-9]{40}$"


def address_schema():
    return str_schema(min_length=42, max_length=42, pattern=ADDRESS_PATTERN)


class Address(HexStr20):
    """
    Use for address-types. Validates as a checksummed address. Left-pads zeroes
    if necessary.
    """

    schema_pattern: ClassVar[str] = ADDRESS_PATTERN
    schema_examples: ClassVar[tuple[str, ...]] = (
        "0x0000000000000000000000000000000000000000",  # Zero address
        "0x02c84e944F97F4A4f60221e6fb5d5DbAE49c7aaB",  # Leading zero
        "0xa5a13f62ce1113838e0d9b4559b8caf5f76463c0",  # Trailing zero
        "0x1e59ce931B4CFea3fe4B875411e280e173cB7A9C",
    )

    @classmethod
    def __get_pydantic_core_schema__(cls, value, handler=None) -> "CoreSchema":
        return with_info_before_validator_function(
            cls.__eth_pydantic_validate__,
            address_schema(),
        )

    @classmethod
    def __eth_pydantic_validate__(
        cls, value: Any, info: ValidationInfo | None = None, **kwargs
    ) -> str:
        value = super().__eth_pydantic_validate__(value, info, pad=PadDirection.LEFT)
        return cls.to_checksum_address(value)

    @classmethod
    def update_schema(cls):
        # Already set statically in the class
        return

    @classmethod
    def to_checksum_address(cls, value: str) -> ChecksumAddress:
        return to_checksum_address(value)


class _AddressTypeFactory:
    @cached_property
    def address_type(self):
        # Lazy define for performance reasons.
        AddressType = Annotated[ChecksumAddress, Address]
        AddressType.__doc__ = """
        A type that can be used in place of ``eth_typing.ChecksumAddress``.

        **NOTE**: We are unable to subclass ``eth_typing.ChecksumAddress``
          in :class:`~eth_pydantic_types.address.Address` because it is
          a NewType; that is why we offer this annotated approach.
        """
        return AddressType


_factory = _AddressTypeFactory()


def __getattr__(name: str):
    if name == "Address":
        return Address

    elif name == "AddressType":
        return _factory.address_type


__all__ = [
    "Address",
    "AddressType",  # noqa: F822
]



from typing import TYPE_CHECKING, Any, ClassVar

from pydantic_core.core_schema import (
    ValidationInfo,
    int_schema,
    no_info_before_validator_function,
    with_info_before_validator_function,
)

from eth_pydantic_types._error import HexValueError
from eth_pydantic_types.hex.base import BaseHex
from eth_pydantic_types.serializers import create_hex_serializer, hex_serializer
from eth_pydantic_types.utils import (
    PadDirection,
    get_hash_examples,
    get_hash_pattern,
    validate_hex_str,
    validate_int_size,
)

if TYPE_CHECKING:
    from pydantic_core import CoreSchema
    from typing_extensions import TypeAlias


class BaseHexInt(int, BaseHex):
    @classmethod
    def __get_pydantic_core_schema__(cls, value, handler=None):
        return no_info_before_validator_function(cls.__eth_pydantic_validate__, int_schema())

    @classmethod
    def __eth_pydantic_validate__(cls, value, **kwargs):
        return value  # Override.

    @classmethod
    def from_bytes(cls, data, byteorder: str = "big", signed: bool = False) -> "BaseHexInt":
        int_value = int.from_bytes(data, byteorder="big", signed=signed)
        return cls(int_value)

    @classmethod
    def validate_hex(cls, data: bytes | str | int):
        if isinstance(data, int):
            return cls(data)

        elif isinstance(data, bytes):
            return cls.from_bytes(data)

        elif isinstance(data, str):
            return cls(int(validate_hex_str(data), 16))

        raise HexValueError(data)

    def __bytes__(self) -> bytes:
        return self.to_bytes(self.size, byteorder="big")


class HexInt(BaseHexInt):
    """A hex int value."""

    @classmethod
    def __get_pydantic_core_schema__(cls, value, handler=None) -> "CoreSchema":
        schema = with_info_before_validator_function(cls.__eth_pydantic_validate__, int_schema())
        schema["serialization"] = hex_serializer
        return schema

    @classmethod
    def __eth_pydantic_validate__(
        cls, value: Any, info: ValidationInfo | None = None, **kwargs
    ) -> int:
        return cls(cls.validate_hex(value))


class BoundHexInt(BaseHexInt):
    """A hex string value, that is required to be a specific size."""

    size: ClassVar[int] = 32
    signed: ClassVar[bool] = False

    @classmethod
    def __get_pydantic_core_schema__(cls, value, handler=None) -> "CoreSchema":
        if cls.signed:
            min_int = -(2 ** (8 * cls.size - 1))
            max_int = 2 ** (8 * cls.size - 1) - 1
        else:
            min_int = 0
            max_int = 2 ** (8 * cls.size) - 1

        schema = with_info_before_validator_function(
            cls.__eth_pydantic_validate__,
            int_schema(le=max_int, ge=min_int),
        )

        # NOTE: Integers should always pad left; else the value increases.
        schema["serialization"] = create_hex_serializer(
            size=cls.size, pad=PadDirection.LEFT, force_even_length=False
        )

        return schema

    @classmethod
    def __eth_pydantic_validate__(
        cls, value: Any, info: ValidationInfo | None = None, **kwargs
    ) -> int:
        hex_int = cls.validate_hex(value)
        sized_value = cls.validate_size(hex_int)
        return cls(sized_value)

    @classmethod
    def validate_size(cls, value: int) -> int:
        cls.update_schema()
        return validate_int_size(value, cls.size, cls.signed)

    @classmethod
    def update_schema(cls):
        str_size = cls.size * 2
        cls.schema_pattern = get_hash_pattern(str_size)
        cls.schema_examples = get_hash_examples(str_size)


class HexInt32(BoundHexInt):
    size: ClassVar[int] = 32


# UInt256 is a common name for this type across other EVM packages, hence this alias.
UInt256: "TypeAlias" = HexInt32



import pytest
from pydantic import BaseModel, ValidationError

from eth_pydantic_types.abi import (
    int8,
    int16,
    int32,
    int64,
    int128,
    int256,
    uint8,
    uint16,
    uint32,
    uint64,
    uint128,
    uint256,
)


class SignedModel(BaseModel):
    valueint8: int8
    valueint16: int16
    valueint32: int32
    valueint64: int64
    valueint128: int128
    valueint256: int256

    @classmethod
    def from_single(cls, value):
        return cls(
            valueint8=value,
            valueint16=value,
            valueint32=value,
            valueint64=value,
            valueint128=value,
            valueint256=value,
        )

    @classmethod
    def from_series(cls, value8, value16, value32, value64, value128, value256):
        return cls(
            valueint8=value8,
            valueint16=value16,
            valueint32=value32,
            valueint64=value64,
            valueint128=value128,
            valueint256=value256,
        )


class UnsignedModel(BaseModel):
    valueuint8: uint8
    valueuint16: uint16
    valueuint32: uint32
    valueuint64: uint64
    valueuint128: uint128
    valueuint256: uint256

    @classmethod
    def from_single(cls, value):
        return cls(
            valueuint8=value,
            valueuint16=value,
            valueuint32=value,
            valueuint64=value,
            valueuint128=value,
            valueuint256=value,
        )

    @classmethod
    def from_series(cls, value8, value16, value32, value64, value128, value256):
        return cls(
            valueuint8=value8,
            valueuint16=value16,
            valueuint32=value32,
            valueuint64=value64,
            valueuint128=value128,
            valueuint256=value256,
        )


def test_negative_unsigned_int():
    with pytest.raises(ValidationError):
        UnsignedModel.from_single(-1)


def test_retains_value():
    signed = SignedModel.from_series(-1, -2, -3, -4, -5, -6)
    assert signed.valueint8 == -1
    assert signed.valueint16 == -2
    assert signed.valueint32 == -3
    assert signed.valueint64 == -4
    assert signed.valueint128 == -5
    assert signed.valueint256 == -6

    unsigned = UnsignedModel.from_series(1, 2, 3, 4, 5, 6)
    assert unsigned.valueuint8 == 1
    assert unsigned.valueuint16 == 2
    assert unsigned.valueuint32 == 3
    assert unsigned.valueuint64 == 4
    assert unsigned.valueuint128 == 5
    assert unsigned.valueuint256 == 6


def test_upper_bounds():
    # Meets the upper bounds
    SignedModel.from_series(
        127,
        32767,
        2147483647,
        9223372036854775807,
        170141183460469231731687303715884105727,
        57896044618658097711785492504343953926634992332820282019728792003956564819967,
    )
    UnsignedModel.from_series(
        255,
        65535,
        4294967295,
        18446744073709551615,
        340282366920938463463374607431768211455,
        115792089237316195423570985008687907853269984665640564039457584007,
    )

    # Exceeds the upper bounds
    with pytest.raises(ValidationError):
        SignedModel.from_series(
            128,
            32768,
            2147483648,
            9223372036854775808,
            170141183460469231731687303715884105728,
            57896044618658097711785492504343953926634992332820282019728792003956564819968,
        )
    with pytest.raises(ValidationError):
        UnsignedModel.from_series(
            256,
            65536,
            4294967296,
            18446744073709551616,
            340282366920938463463374607431768211456,
            115792089237316195423570985008687907853269984665640564039457584008,
        )


def test_lower_bounds():
    # Meets the lower bounds
    SignedModel.from_series(
        -128,
        -32768,
        -2147483648,
        -9223372036854775808,
        -170141183460469231731687303715884105727,
        -57896044618658097711785492504343953926634992332820282019728792003956564819968,
    )
    UnsignedModel.from_series(0, 0, 0, 0, 0, 0)

    # Exceeds the lower bounds
    with pytest.raises(ValidationError):
        SignedModel.from_series(
            -129,
            -32769,
            -2147483649,
            -9223372036854775809,
            -170141183460469231731687303715884105728,
            -57896044618658097711785492504343953926634992332820282019728792003956564819969,
        )
    with pytest.raises(ValidationError):
        UnsignedModel.from_series(-1, -1, -1, -1, -1, -1)



import pytest
from pydantic import BaseModel, ValidationError, field_validator

from eth_pydantic_types.hex import (
    HexBytes,
    HexStr,
    HexStr20,
)
from eth_pydantic_types.utils import PadDirection


class StrModel(BaseModel):
    value: HexStr


class TestHexStr:
    @pytest.mark.parametrize("value", ("0xa", 10, HexBytes(10)))
    def test_valid(self, value):
        actual = StrModel(value=value)

        # The end result, the value is a str
        assert actual.value == "0x0a"
        assert isinstance(actual.value, str)

    def test_invalid(self):
        with pytest.raises(ValidationError):
            StrModel(value="foo")

    def test_conversions(self):
        model = StrModel(value="0x123")
        assert int(model.value, 16) == 291
        assert bytes.fromhex(model.value[2:]) == b"\x01#"

    def test_schema(self):
        actual = StrModel.model_json_schema()
        properties = actual["properties"]
        assert len(properties) == 1
        prop = properties["value"]
        assert prop["type"] == "string"
        assert prop["format"] == "binary"
        assert prop["pattern"] == "^0x([0-9a-f][0-9a-f])*$"
        assert prop["examples"] == [
            "0x",
            "0xd4",
            "0xd4e5",
            "0xd4e56740",
            "0xd4e56740f876aef8",
            "0xd4e56740f876aef8c010b86a40d5f567",
            "0xd4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3",
        ]

    def test_model_dump(self, bytes32str):
        model = StrModel(value=bytes32str)
        actual = model.model_dump()
        expected = {"value": "0x9b70bd98ccb5b6434c2ead14d68d15f392435a06ff469f8d1f8cf38b2ae0b0e2"}
        assert actual == expected

        model = StrModel(value=3)
        actual = model.model_dump()
        expected = {"value": "0x03"}
        assert actual == expected

    def test_from_bytes(self):
        value = b"\xb7\xfc\xef\x7f\xe7E\xf2\xa9U`\xff_U\x0e;\x8f"
        actual = HexStr.from_bytes(value)
        assert actual.startswith("0x")

    def test_hex_removes_leading_zeroes_if_needed(self):
        address = "0x000000000000000000000000cafac3dd18ac6c6e92c921884f9e4176737c052c"

        class MyModel(BaseModel):
            my_address: HexStr20

        # Test both str and bytes for input.
        for addr in (address, HexBytes(address)):
            model = MyModel(my_address=addr)
            assert len(model.my_address) == 42
            assert model.my_address == "0xcafac3dd18ac6c6e92c921884f9e4176737c052c"

        def test_from_bytes(self):
            value = b"\x101\xf0\xc9\xacT\xdc\xb6KO\x12\x1a'\x95|\x14&<\\\xb4"
            model = StrModel(value=value)
            assert model.value == HexBytes(value)

    def test_right_pad(self):
        class MyModel(BaseModel):
            my_str: HexStr20

            @field_validator("my_str", mode="before")
            @classmethod
            def validate_value(cls, value, info):
                field_type = cls.model_fields[info.field_name].annotation
                assert field_type  # For Mypy
                return field_type.__eth_pydantic_validate__(value, pad=PadDirection.RIGHT)

        model = MyModel(my_str=1)
        assert model.my_str.startswith("0x01")



import pytest
from pydantic import BaseModel, ValidationError

from eth_pydantic_types.hex import (
    HexBytes,
    HexBytes20,
    HexBytes32,
    HexStr32,
)


class SizedModel(BaseModel):
    valuebytes20: HexBytes20
    valuebytes32: HexBytes32
    valuestr32: HexStr32

    @classmethod
    def from_single(cls, value):
        return cls(
            valuebytes20=value,
            valuebytes32=value,
            valuestr32=value,
        )


class TestSized:
    """
    Testing all or any sized hex types in a pydantic model.
    """

    @pytest.mark.parametrize("value", ("0x32", HexBytes("0x32"), b"2", 50))
    def test_valid(self, value):
        model = SizedModel.from_single(value)
        assert len(model.valuebytes20) == 20
        assert len(model.valuebytes32) == 32
        assert len(model.valuestr32) == 66

        if isinstance(value, int):
            assert model.valuebytes20.hex().endswith("32")
            assert model.valuebytes32.hex().endswith("32")
            assert model.valuestr32.endswith("32")
        else:
            assert model.valuebytes20.hex().startswith("32")
            assert model.valuebytes32.hex().startswith("32")
            assert model.valuestr32.startswith("0x32")

    @pytest.mark.parametrize("value", ("foo", -35, "0x" + ("F" * 100)))
    def test_invalid_hash(self, value):
        with pytest.raises(ValidationError):
            SizedModel.from_single(value)

    def test_construct_removes_leading_zeroes_if_needed(self):
        address = "0x000000000000000000000000cafac3dd18ac6c6e92c921884f9e4176737c052c"

        class MyModel(BaseModel):
            my_address: HexBytes20

        # Test both str and bytes for input.
        for addr in (address, HexBytes(address)):
            model = MyModel(my_address=addr)
            assert len(model.my_address) == 20
            assert model.my_address == HexBytes("0xcafac3dd18ac6c6e92c921884f9e4176737c052c")

    def test_schema(self):
        actual = SizedModel.model_json_schema()
        for name, prop in actual["properties"].items():
            is_bytes = "bytes" in name
            if is_bytes:
                size_from_name = int(name.replace("valuebytes", ""))
                expected_length = size_from_name
            else:
                size_from_name = int(name.replace("valuestr", ""))
                expected_length = size_from_name * 2 + 2

            hex_value_str_size = size_from_name * 2

            assert prop["maxLength"] == expected_length
            assert prop["minLength"] == expected_length
            assert prop["type"] == "string"
            assert prop["format"] == "binary"
            assert prop["pattern"] == f"^0x[a-fA-F0-9]{{{hex_value_str_size}}}$"

            assert prop["examples"][0] == f"0x{'0' * hex_value_str_size}"
            assert prop["examples"][1].startswith("0x0")  # Leading zero
            assert prop["examples"][2].endswith("0")  # Trailing zero
            assert all(len(ex) == hex_value_str_size + 2 for ex in prop["examples"])

    def test_model_dump(self, bytes32str):
        model = SizedModel.from_single(5)
        actual = model.model_dump()
        expected = {
            "valuebytes20": "0x0000000000000000000000000000000000000005",
            "valuebytes32": "0x0000000000000000000000000000000000000000000000000000000000000005",
            "valuestr32": "0x0000000000000000000000000000000000000000000000000000000000000005",
        }
        assert actual == expected

    def test_right_versus_left_pad(self):
        class SimpleModel(BaseModel):
            valuebytes: HexBytes32
            valuestr: HexStr32

        leftpad = SimpleModel(valuebytes=5, valuestr=5)
        rightpad = SimpleModel(valuebytes="0x05", valuestr="0x05")
        assert leftpad.valuebytes == HexBytes(
            "0x0000000000000000000000000000000000000000000000000000000000000005"
        )
        assert leftpad.valuestr == (
            "0x0000000000000000000000000000000000000000000000000000000000000005"
        )
        assert rightpad.valuebytes == HexBytes(
            "0x0500000000000000000000000000000000000000000000000000000000000000"
        )
        assert rightpad.valuestr == (
            "0x0500000000000000000000000000000000000000000000000000000000000000"
        )



from typing import ClassVar

import pytest
from pydantic import BaseModel, ValidationError

from eth_pydantic_types.hex import (
    BoundHexInt,
    HexInt,
    HexInt32,
    UInt256,
)
from eth_pydantic_types.hex.bytes import HexBytes


class IntModel(BaseModel):
    value: HexInt


class HexInt32Model(BaseModel):
    value: HexInt32


class UInt256Model(BaseModel):
    value: UInt256


class TestHexInt:
    @pytest.mark.parametrize("value", ("0xa", 10, b"\n"))
    def test_valid(self, value):
        actual = IntModel(value=value)
        assert actual.value == 10

    def test_invalid(self):
        with pytest.raises(ValidationError):
            _ = IntModel(value="foo")

    def test_schema(self):
        actual = IntModel.model_json_schema()
        assert actual["properties"]["value"]["type"] == "integer"
        assert actual["properties"]["value"]["examples"][0].startswith("0x")

    @pytest.mark.parametrize("value", ("0xa", "0x0a", 10))
    def test_model_dump(self, value):
        model = IntModel(value=value)
        actual = model.model_dump()

        # Notice the odd number of bytes here; that is expected for HexInt.
        # It is less storage; we know what number it is.
        assert actual == {"value": "0xa"}

    @pytest.mark.parametrize("value", (-1, 2 * 8**256))
    def test_out_of_bounds(self, value):
        class PublicKey(BoundHexInt):
            size: ClassVar[int] = 64
            signed: ClassVar[bool] = False

        class Model(BaseModel):
            pub_key: PublicKey

        with pytest.raises(ValidationError):
            _ = Model(pub_key=value)


class TestHexInt32:
    @pytest.mark.parametrize("value", ("0xa", 10, b"\n"))
    def test_valid(self, value):
        actual = HexInt32Model(value=value)
        assert actual.value == 10

        str_value = actual.model_dump()["value"]
        expected = "0x000000000000000000000000000000000000000000000000000000000000000a"
        assert str_value == expected

        # The resulting size in bytes is 32.
        assert len(HexBytes(str_value)) == 32


# Even though UInt256 is a TypeAlias of HexInt32, we want to ensure it functions separately.
class TestUInt256:
    @pytest.mark.parametrize("value", ("0xa", 10, b"\n"))
    def test_valid(self, value):
        actual = UInt256Model(value=value)
        assert actual.value == 10

        str_value = actual.model_dump()["value"]
        expected = "0x000000000000000000000000000000000000000000000000000000000000000a"
        assert str_value == expected

        # The resulting size in bytes is 32.
        assert len(HexBytes(str_value)) == 32


def test_custom_type():
    # An 8-byte unsigned hex integer.
    class NetworkID(BoundHexInt):
        size: int = 8
        signed: bool = False

    class MyModel(BaseModel):
        network_id: NetworkID

    # Sepolia.
    model = MyModel(network_id=11155111)
    assert model.network_id == 11155111

    model_json = model.model_dump_json()
    assert model_json == '{"network_id":"0x0000000000aa36a7"}'



from collections.abc import Sized
from enum import Enum
from typing import TYPE_CHECKING, Callable, TypeVar

from eth_pydantic_types._error import HexValueError, SizeError

if TYPE_CHECKING:
    __SIZED_T = TypeVar("__SIZED_T", bound=Sized)


class PadDirection(str, Enum):
    LEFT = "left"
    RIGHT = "right"


def validate_size(value: "__SIZED_T", size: int, coerce: Callable | None = None) -> "__SIZED_T":
    if len(value) == size:
        return value

    elif coerce:
        return validate_size(coerce(value), size)

    raise SizeError(size, value)


def validate_in_range(value: int, size: int, signed: bool = True) -> int:
    if signed:
        if -(2**size) / 2 <= value < (2**size) / 2:
            return value

    else:
        if 0 <= value < 2 ** (size * 8):
            return value

    raise SizeError(size, value)


def validate_bytes_size(
    value: bytes, size: int, pad_direction: PadDirection = PadDirection.LEFT
) -> bytes:
    return validate_size(
        value,
        size,
        coerce=lambda v: _coerce_hexbytes_size(v, size, pad_direction=pad_direction),
    )


def validate_address_size(value: str) -> str:
    return validate_str_size(value, 40)


def validate_str_size(
    value: str, size: int, pad_direction: PadDirection = PadDirection.LEFT
) -> str:
    return validate_size(
        value,
        size,
        coerce=lambda v: _coerce_hexstr_size(v, size, pad_direction=pad_direction),
    )


def validate_int_size(value: int, size: int, signed: bool) -> int:
    return validate_in_range(value, size, signed)


def _coerce_hexstr_size(
    val: str, length: int, pad_direction: PadDirection = PadDirection.LEFT
) -> str:
    val = val.replace("0x", "") if val.startswith("0x") else val
    if len(val) == length:
        return val

    val_stripped = val.lstrip("0")

    # Ensure is still even.
    if len(val_stripped) % 2 != 0:
        val_stripped = f"0{val_stripped}"

    # Put the remaining zeroes on the side we said to pad.
    num_zeroes = max(0, length - len(val_stripped))
    zeroes = "0" * num_zeroes
    return (
        f"{zeroes}{val_stripped}"
        if pad_direction is PadDirection.LEFT
        else f"{val_stripped}{zeroes}"
    )


def _coerce_hexbytes_size(
    val: bytes, num_bytes: int, pad_direction: PadDirection = PadDirection.LEFT
) -> bytes:
    if len(val) == num_bytes:
        return val

    val_stripped = val.lstrip(b"\x00")
    num_zeroes = max(0, num_bytes - len(val_stripped))
    zeroes = b"\x00" * num_zeroes
    return zeroes + val_stripped if pad_direction is PadDirection.LEFT else val_stripped + zeroes


def validate_hex_str(value: str) -> str:
    hex_value = (value[2:] if value.startswith("0x") else value).lower()
    if set(hex_value) - set("1234567890abcdef"):
        raise HexValueError(value)

    # Missing zero padding.
    if len(hex_value) % 2 != 0:
        hex_value = f"0{hex_value}"

    return f"0x{hex_value}"


def get_hash_pattern(str_size: int) -> str:
    return f"^0x[a-fA-F0-9]{{{str_size}}}$"


def get_hash_examples(str_size: int) -> tuple[str, str, str, str]:
    zero_hash = f"0x{'0' * str_size}"
    leading_zero = f"0x01{'1e' * ((str_size - 1) // 2)}"
    trailing_zero = f"0x{'1e' * ((str_size - 1) // 2)}10"
    full_hash = f"0x{'1e' * (str_size // 2)}"
    return zero_hash, leading_zero, trailing_zero, full_hash



from typing import ClassVar

import pytest
from eth_utils import to_hex
from hexbytes import HexBytes as BaseHexBytes
from pydantic import BaseModel, ValidationError, field_validator

from eth_pydantic_types.hex import BoundHexBytes, HexBytes, HexBytes32
from eth_pydantic_types.hex.bytes import HexBytes20
from eth_pydantic_types.utils import PadDirection


class BytesModel(BaseModel):
    value: HexBytes


class TestHexBytes:
    @pytest.mark.parametrize("value", ("0xa", 10, b"\n"))
    def test_valid(self, value):
        actual = BytesModel(value=value)

        # The end result, the value is a hexbytes.HexBytes
        assert actual.value == BaseHexBytes(value)
        assert to_hex(actual.value) == "0x0a"
        assert isinstance(actual.value, bytes)
        assert isinstance(actual.value, BaseHexBytes)

    def test_invalid(self):
        with pytest.raises(ValidationError):
            BytesModel(value="foo")

    def test_fromhex(self, bytes32str):
        actual_with_0x = HexBytes.fromhex(bytes32str)
        actual_without_0x = HexBytes.fromhex(bytes32str[2:])
        expected = HexBytes(bytes32str)
        assert actual_with_0x == actual_without_0x == expected

    def test_schema(self):
        actual = BytesModel.model_json_schema()
        prop = actual["properties"]["value"]
        assert prop["type"] == "string"
        assert prop["format"] == "binary"
        assert prop["pattern"] == "^0x([0-9a-f][0-9a-f])*$"
        assert prop["examples"] == [
            "0x",
            "0xd4",
            "0xd4e5",
            "0xd4e56740",
            "0xd4e56740f876aef8",
            "0xd4e56740f876aef8c010b86a40d5f567",
            "0xd4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3",
        ]

    def test_model_dump(self, bytes32str):
        model = BytesModel(value=bytes32str)
        actual = model.model_dump()
        expected = {"value": "0x9b70bd98ccb5b6434c2ead14d68d15f392435a06ff469f8d1f8cf38b2ae0b0e2"}
        assert actual == expected

    def test_from_bytes(self):
        value = b"\x101\xf0\xc9\xacT\xdc\xb6KO\x12\x1a'\x95|\x14&<\\\xb4"
        model = BytesModel(value=value)
        assert model.value == HexBytes(value)


class TestHexBytes32:
    def test_fromhex(self, bytes32str):
        actual_with_0x = HexBytes32.fromhex(bytes32str)
        actual_without_0x = HexBytes32.fromhex(bytes32str[2:])
        expected = HexBytes(bytes32str)
        assert actual_with_0x == actual_without_0x == expected

    def test_is_bytes(self, bytes32str):
        assert isinstance(HexBytes32.fromhex(bytes32str), bytes)


class TestBoundHexBytes:
    """
    Showing how to use BoundHexBytes to make your own bounded-types.
    """

    def test_use_bound_type(self):
        class PublicKey(BoundHexBytes):
            size: ClassVar[int] = 64

        class Account(BaseModel):
            pub_key: PublicKey

        key = (
            "0x000000000000000000000000000000000000000000000000000000000000000"
            "0000000000000000000000000000000000000000000000000000000000000005"
        )
        account = Account(pub_key=key)
        assert account.pub_key == HexBytes(key)

    def test_right_pad(self):
        class MyModel(BaseModel):
            my_bytes: HexBytes20

            @field_validator("my_bytes", mode="before")
            @classmethod
            def validate_value(cls, value, info):
                return HexBytes20.__eth_pydantic_validate__(value, pad=PadDirection.RIGHT)

        model = MyModel(my_bytes=1)
        assert model.my_bytes.startswith(HexBytes(1))



from .address import Address, AddressType
from .bip122 import Bip122Uri
from .hex import (
    BoundHexBytes,
    BoundHexStr,
    HexBytes,
    HexBytes20,
    HexBytes32,
    HexStr,
    HexStr20,
    HexStr32,
)

__all__ = [
    "Address",
    "AddressType",
    "Bip122Uri",
    "BoundHexBytes",
    "BoundHexStr",
    "HexBytes",
    "HexBytes20",
    "HexBytes32",
    "HexStr",
    "HexStr20",
    "HexStr32",
]
